
#include <esp_timer.h>

#include <lv_init.h>

#include <mutex>

#include "display.h"

// TODO: Remove this dependency by relocating SSD1306::oledb_buffer_
#include "ssd1306.h"

// LVGL library is not thread-safe, this example calls LVGL APIs from tasks.
// We must use a mutex to protect it.
_lock_t Display::ScopedLock::lv_lock_;

// Static TimeKeeper for managing ESP timers across all displays.
TimeKeeper Display::timers_;

Display::Display(IPanelDevice &device) :
    panel_(device),
    lv_buf_(nullptr)
{
  if (!lv_is_initialized()) {
    ESP_LOGI(TAG, "Initialize LVGL");
    lv_init();
  }

  ESP_LOGI(TAG, "Creating LVGL display");
  lv_display_ = panel_.device_->create_display();
  // associate the i2c panel handle to the display
  lv_display_set_user_data(lv_display_, panel_.esp_panel_);

  register_draw_buffer();
  register_lvgl_tick_timer();

  ESP_LOGI(TAG, "Create LVGL FreeRTOS task");
  xTaskCreate(Display::lvgl_port_task, "LVGL", LVGL_TASK_STACK_SIZE,
              nullptr, LVGL_TASK_PRIORITY, nullptr);
}

void Display::set_text(const char *text,
                       const char *name,
                       lv_label_long_mode_t long_mode,
                       lv_align_t align)
{
  // Lock the mutex due to the LVGL APIs are not thread-safe.
  ScopedLock lock;

  ESP_LOGI(TAG, "Display LVGL Scroll Text");
  lv_obj_t *scr = lv_display_get_screen_active(lv_display_);

  // Create the label if it's `name` doesn't already exist in the map keys.
  if (!lv_objects_.count(name)) {
    lv_objects_[name] = lv_label_create(scr);
  }
  auto obj = lv_objects_[name];

  // Circular scroll.
  lv_label_set_long_mode(obj, long_mode);
  lv_label_set_text(obj, text);

  // Set the size of the screen.
  // If you use rotation 90 or 270 use lv_display_get_vertical_resolution.
  lv_obj_set_width(obj, lv_display_get_horizontal_resolution(lv_display_));
  lv_obj_align(obj, align, 0, 0);
}

bool Display::lvgl_flush_ready_cb(esp_lcd_panel_io_handle_t,
                                  esp_lcd_panel_io_event_data_t *,
                                  void *user_ctx)
{
  auto *disp = (lv_display_t *) user_ctx;
  lv_display_flush_ready(disp);
  return false;
}

void Display::lvgl_flush_cb(lv_display_t *display, const lv_area_t *area,
                            uint8_t *px_map) // NOLINT(*-non-const-parameter)
{
  auto panel_handle =
      (esp_lcd_panel_handle_t) lv_display_get_user_data(display);

  // Necessary because LVGL reserves 2x4 bytes in the buffer for a palette.
  // For more information about the monochrome, please refer to:
  // https://docs.lvgl.io/9.2/porting/display.html#monochrome-displays
  // Skip the palette here.
  px_map += LVGL_PALETTE_SIZE;

  uint16_t hor_res = lv_display_get_physical_horizontal_resolution(display);
  int32_t x1 = area->x1;
  int32_t x2 = area->x2;
  int32_t y1 = area->y1;
  int32_t y2 = area->y2;

  for (int32_t y = y1; y <= y2; y++) {
    for (int32_t x = x1; x <= x2; x++) {
      /* The order of bits is MSB first.
                  MSB           LSB
         bits      7 6 5 4 3 2 1 0
         pixels    0 1 2 3 4 5 6 7
                  Left         Right
      */
      bool chroma_color = (px_map[(hor_res >> 3) * y + (x >> 3)] &
                           1 << (7 - x % 8));

      // Write to the buffer as required for the display.
      // It writes only 1-bit for monochrome displays mapped vertically.
      uint8_t *buf = SSD1306::oled_buffer_ + hor_res * (y >> 3) + (x);
      if (chroma_color) {
        (*buf) &= ~(1 << (y % 8));
      } else {
        (*buf) |= (1 << (y % 8));
      }
    }
  }
  // Pass the draw buffer to the driver.
  ESP_ERROR_CHECK(
      esp_lcd_panel_draw_bitmap(panel_handle, x1, y1, x2 + 1, y2 + 1,
                                SSD1306::oled_buffer_));
}

void Display::lvgl_increase_tick_cb(void *)
{
  // Tell LVGL how many milliseconds has elapsed
  lv_tick_inc(LVGL_TICK_PERIOD_MS);
}

[[noreturn]] void Display::lvgl_port_task(void *)
{
  ESP_LOGI(TAG, "Starting LVGL task");
  for (uint32_t time_to_next_ms = 0; true; usleep(1000 * time_to_next_ms)) {
    ScopedLock lock;
    time_to_next_ms = lv_timer_handler();
  }
}

void Display::register_draw_buffer()
{
  // Create draw buffer.
  ESP_LOGI(TAG, "Allocate separate LVGL draw buffers");
  lv_buf_ = heap_caps_calloc(1, panel_.device_->lv_buf_size_,
                             MALLOC_CAP_INTERNAL | MALLOC_CAP_8BIT);
  assert(lv_buf_);

  ESP_LOGI(TAG, "Set LVGL draw buffers");
  // Color format must be set first, LVGL9 suooprt new monochromatic format.
  lv_display_set_color_format(lv_display_, LV_COLOR_FORMAT_I1);
  lv_display_set_buffers(lv_display_, lv_buf_, nullptr,
                         panel_.device_->lv_buf_size_,
                         LV_DISPLAY_RENDER_MODE_FULL);
  lv_display_set_rotation(lv_display_, LV_DISPLAY_ROTATION_0);

  ESP_LOGI(TAG, "Set LVGL callback for flushing to the display");
  lv_display_set_flush_cb(lv_display_, Display::lvgl_flush_cb);

  ESP_LOGI(TAG, "Register io panel callback for LVGL flush ready notification");
  const esp_lcd_panel_io_callbacks_t cbs = {
      .on_color_trans_done = Display::lvgl_flush_ready_cb,
  };
  ESP_ERROR_CHECK(
      esp_lcd_panel_io_register_event_callbacks(panel_.esp_io_, &cbs,
                                                lv_display_));
}

void Display::register_lvgl_tick_timer()
{
  ESP_LOGI(TAG, "Use esp_timer to increase LVGL tick");
  const esp_timer_create_args_t esp_timer_args = {
      .callback = &Display::lvgl_increase_tick_cb,
      .name = "lvgl_tick"
  };
  timers_.start_new_timer_periodic(esp_timer_args, LVGL_TICK_PERIOD_MS * 1000);
}
